"""BoardGameGeek JSON spider."""

import csv
import os
from datetime import timezone
from io import StringIO
from pathlib import Path
from typing import Optional

from pytility import parse_date, parse_int
from scrapy import Request, Spider

from ..items import GameItem
from ..utils import extract_query_param, json_from_response, now

BASE_DIR = Path(__file__).parent.parent.parent.parent.resolve()
GAMES_FILE = (BASE_DIR / "board-game-data" / "scraped" / "bgg_GameItem.csv").resolve()


class BggJsonSpider(Spider):
    """BoardGameGeek JSON spider."""

    name = "bgg_json_rankings"
    allowed_domains = ("geekdo.com",)
    start_urls = (GAMES_FILE.as_uri(),)
    item_classes = (GameItem,)

    url = (
        "https://api.geekdo.com/api/historicalrankgraph"
        + "?objectid={item_id}&objecttype=thing&rankobjectid={game_type_id}"
    )
    game_types = {
        "overall": 1,
        "abstract": 4666,
        "children": 4665,
        "customizable": 4667,
        "family": 5499,
        "party": 5498,
        "strategy": 5497,
        "thematic": 5496,
        "war": 4664,
    }
    id_field = "bgg_id"

    custom_settings = {
        "DOWNLOAD_DELAY": 10.0,
        "CONCURRENT_REQUESTS_PER_DOMAIN": 4,
        "AUTOTHROTTLE_TARGET_CONCURRENCY": 2,
        "DELAYED_RETRY_ENABLED": True,
        "DELAYED_RETRY_HTTP_CODES": (202,),
        "DELAYED_RETRY_DELAY": 30.0,
        "AUTOTHROTTLE_HTTP_CODES": (429, 503, 504),
        "ITEM_PIPELINES": {"scrapy_extensions.ValidatePipeline": None},
    }

    def get_game_type(self) -> str:
        """Get the game type from settings."""
        return (
            getattr(self, "game_type", None)
            or self.settings.get("GAME_TYPE")
            or os.getenv("GAME_TYPE")
            or "overall"
        )

    def get_game_type_id(self, game_type: Optional[str] = None) -> Optional[int]:
        """Get the object ID corresponding to the game type."""
        game_type = game_type or self.get_game_type()
        return self.game_types.get(game_type)

    def parse_csv(self, text, id_field=None):
        """Parse a CSV string for IDs."""
        id_field = id_field or self.id_field
        file = StringIO(text, newline="")
        for game in csv.DictReader(file):
            item_id = parse_int(game.get(id_field))
            if item_id:
                yield item_id, game.get("name"), parse_int(game.get("num_votes"))

    def parse(self, response):
        """
        @url file:///Users/markus/Recommend.Games/board-game-data/scraped/bgg_GameItem.csv
        @returns items 0 0
        @returns requests 100000
        """

        game_type = self.get_game_type()
        game_type_id = self.get_game_type_id(game_type)

        if not game_type_id:
            self.logger.error("Invalid game type <%s>, aborting", game_type)
            return

        self.logger.info(
            "Scraping rankings for game type <%s> (ID %d)",
            game_type,
            game_type_id,
        )

        try:
            for item_id, name, priority in self.parse_csv(response.text):
                meta = {"name": name, "item_id": item_id}
                yield Request(
                    url=self.url.format(game_type_id=game_type_id, item_id=item_id),
                    callback=self.parse_game,
                    meta=meta,
                    priority=priority or 0,
                )

        except Exception:
            self.logger.exception(
                "Response <%s> cannot be processed as CSV",
                response.url,
            )

    def parse_game(self, response):
        # pylint: disable=line-too-long
        """
        @url https://api.geekdo.com/api/historicalrankgraph?objectid=13&objecttype=thing&rankobjectid=1
        @returns items 5000
        @returns requests 0 0
        @scrapes bgg_id rank published_at scraped_at
        """

        result = json_from_response(response)
        data = result.get("data") or ()

        name = response.meta.get("name")
        item_id = parse_int(response.meta.get("item_id")) or parse_int(
            extract_query_param(response.url, "objectid")
        )

        if not item_id:
            self.logger.warning(
                "Unable to extract item ID from <%s>, skipping…",
                response.url,
            )
            return

        if not data:
            self.logger.debug(
                "No data for item ID <%d>, consider removing the request",
                item_id,
            )
            return

        scraped_at = now()

        for date, rank in data:
            published_at = parse_date(date / 1000, tzinfo=timezone.utc)
            yield GameItem(
                name=name,
                bgg_id=item_id,
                rank=rank,
                published_at=published_at,
                scraped_at=scraped_at,
            )



"""Download the latest BGG data dump."""

import argparse
import logging
import os
import sys

from pathlib import Path
from typing import Union

import requests

from scrapy.selector import Selector

BASE_DIR = Path(__file__).resolve().parent.parent
LOGGER = logging.getLogger(__name__)


def download_bgg_dump(
    username: str,
    password: str,
    target_dir: Union[str, Path],
    force_overwrite: bool = False,
) -> None:
    """Download the latest BGG data dump."""

    target_dir = Path(target_dir).resolve()
    target_dir.mkdir(parents=True, exist_ok=True)

    LOGGER.info("Downloading latest BGG dump to <%s>", target_dir)

    login_url = "https://boardgamegeek.com/login/api/v1"
    html_url = "https://boardgamegeek.com/data_dumps/bg_ranks"

    with requests.Session() as session:
        credentials = {
            "credentials": {
                "username": username,
                "password": password,
            }
        }
        login_response = session.post(login_url, json=credentials)
        login_response.raise_for_status()

        html_response = session.get(html_url)
        html_response.raise_for_status()

        selector = Selector(text=html_response.text)

        for link in selector.css("#maincontent a[download]"):
            download_url = link.xpath("@href").get()
            file_name = link.xpath("@download").get()
            file_path = target_dir / file_name

            LOGGER.info("Downloading <%s> to <%s>", download_url, file_path)

            if file_path.exists():
                if not force_overwrite:
                    LOGGER.info("Skipping <%s> as it already exists.", file_path)
                    continue
                LOGGER.info("Overwriting <%s> as it already exists.", file_path)

            download_response = session.get(download_url)
            download_response.raise_for_status()

            with file_path.open("wb") as file:
                file.write(download_response.content)

    LOGGER.info("Done.")


def _parse_args():
    parser = argparse.ArgumentParser(description="Download the latest BGG data dump.")
    parser.add_argument(
        "--out-dir",
        "-d",
        default=BASE_DIR / "feeds" / "bgg_dump",
        help="Output directory",
    )
    parser.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="Force overwrite existing files",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="Log level (repeat for more verbosity)",
    )

    return parser.parse_args()


def main():
    """Command line entry point."""

    args = _parse_args()

    logging.basicConfig(
        stream=sys.stderr,
        level=logging.DEBUG if args.verbose > 0 else logging.INFO,
        format="%(asctime)s %(levelname)-8.8s [%(name)s:%(lineno)s] %(message)s",
    )

    LOGGER.info(args)

    username = os.getenv("BGG_USERNAME")
    password = os.getenv("BGG_PASSWORD")

    download_bgg_dump(
        username=username,
        password=password,
        target_dir=args.out_dir,
        force_overwrite=args.force,
    )


if __name__ == "__main__":
    main()
